/*
** pork_buddy.c - Buddy list / permit / deny management.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <pork.h>

static int buddy_hash_compare(void *l, void *r) {
	char *str = l;
	struct buddy *buddy = r;

	return (strcasecmp(str, buddy->nname));
}

static void buddy_hash_cleanup(void *data) {
	struct buddy *buddy = data;

	free(buddy->nname);
	free(buddy->name);
	free(buddy);
}

static void group_cleanup(void *data) {
	struct group *group = data;

	free(group->name);
	dlist_destroy(group->member_list,  NULL);
}

static void generic_cleanup(void *data) {
	free(data);
}

static int buddy_rights_compare(void *l, void *r) {
	char *str1 = l;
	char *str2 = r;

	return (strcasecmp(str1, str2));
}

static int group_compare(void *l, void *r) {
	char *name = l;
	struct group *gr = r;

	return (strcasecmp(name, gr->name));
}

static int buddy_hash_remove(struct buddy_pref *pref, const char *screen_name) {
	char buddy[NUSER_LEN];
	u_int32_t hash;

	normalize(screen_name, buddy, sizeof(buddy));
	hash = string_hash(buddy, pref->buddy_hash.order);

	return (hash_remove(&pref->buddy_hash, buddy, hash));
}

static char *buddy_list_file(const char *username) {
	char *pork_dir;
	char *path;
	size_t len;

	pork_dir = opt_get_str(OPT_PORK_DIR);

	len = strlen(pork_dir) + strlen(username) + sizeof(".blist") + 2;
	path = xmalloc(len);

	snprintf(path, len, "%s/%s.blist", pork_dir, username);
	return (path);
}

int buddy_remove(struct aim_priv *priv, const char *screen_name) {
	struct buddy *buddy;
	dlist_t *list_node;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	normalize(screen_name, nname, sizeof(nname));
	buddy = buddy_find(priv, nname);
	if (buddy == NULL)
		return (-1);

	list_node = dlist_find(buddy->group->member_list,
					buddy->nname, buddy_hash_compare);

	if (list_node == NULL)
		return (-1);

	aim_buddy_remove(priv, buddy);
	buddy->group->member_list = dlist_remove(buddy->group->member_list,
									list_node);
	buddy->group->num_members--;
	blist_del(priv->blist, buddy);
	blist_draw_group_label(priv->blist, buddy->group);

	return (buddy_hash_remove(pref, buddy->name));
}

/*
** Add the buddy whose screen name is "screen_name" to the specified
** group.
*/

struct buddy *buddy_add(struct aim_priv *priv,
						const char *screen_name,
						struct group *group)
{
	struct buddy *buddy;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	normalize(screen_name, nname, sizeof(nname));
	if (buddy_find(priv, nname) != NULL)
		return (NULL);

	buddy = xcalloc(1, sizeof(*buddy));

	buddy->name = xstrdup(screen_name);
	buddy->nname = xstrdup(nname);
	buddy->group = group;
	buddy->refnum = group->high_refnum++;

	group->num_members++;
	group->member_list = dlist_add_tail(group->member_list, buddy);

	hash_add(&pref->buddy_hash, buddy,
		string_hash(buddy->nname, pref->buddy_hash.order));

	aim_buddy_add(priv, buddy);
	blist_draw_group_label(priv->blist, group);
	return (buddy);
}

struct group *group_find(struct aim_priv *priv, char *group_name) {
	dlist_t *node;
	struct buddy_pref *pref = priv->buddy_pref;

	node = dlist_find(pref->group_list, group_name, group_compare);
	if (node == NULL)
		return (NULL);

	return (node->data);
}

struct group *group_add(struct aim_priv *priv, char *group_name) {
	struct group *group;
	struct buddy_pref *pref = priv->buddy_pref;

	group = group_find(priv, group_name);
	if (group != NULL)
		return (group);

	group = xcalloc(1, sizeof(*group));
	group->name = xstrdup(group_name);
	group->refnum = pref->next_group_refnum++;
	pref->group_list = dlist_add_tail(pref->group_list, group);

	blist_add_group(priv->blist, group);
	/* XXX: add the SSI group item */

	return (group);
}

int group_remove(struct aim_priv *priv, char *group_name) {
	struct group *group;
	dlist_t *memb_list;
	dlist_t *node;
	struct buddy_pref *pref = priv->buddy_pref;

	node = dlist_find(pref->group_list, group_name, group_compare);
	if (node == NULL)
		return (-1);

	group = node->data;
	memb_list = group->member_list;

	while (memb_list != NULL) {
		dlist_t *next = memb_list->next;
		struct buddy *buddy = memb_list->data;

		buddy_remove(priv, buddy->name);
		memb_list = next;
	}

	pref->group_list = dlist_remove(pref->group_list, node);

	blist_del_group(priv->blist, group);
	/* XXX - remove it from aim ssi list */

	return (0);
}

struct buddy *buddy_find(struct aim_priv *priv, const char *screen_name) {
	dlist_t *node;
	char buddy[NUSER_LEN];
	u_int32_t hash;
	struct buddy_pref *pref = priv->buddy_pref;

	normalize(screen_name, buddy, sizeof(buddy));
	hash = string_hash(buddy, pref->buddy_hash.order);

	node = hash_find(&pref->buddy_hash, buddy, hash);
	if (node != NULL)
		return (node->data);

	return (NULL);
}

void buddy_alias(struct aim_priv *priv, struct buddy *buddy, char *alias) {
	free(buddy->name);
	buddy->name = xstrdup(alias);

	if (buddy->blist_line != NULL)
		blist_draw_buddy_label(priv->blist, buddy);
}

int buddy_init(struct aim_priv *priv) {
	struct buddy_pref *pref;
	char *buddy_list_fn;

	if (priv->buddy_pref != NULL)
		return (-1);

	if (priv->username == NULL)
		return (-1);

	pref = xcalloc(1, sizeof(*pref));
	priv->buddy_pref = pref;

	hash_init(&pref->buddy_hash, 5, buddy_hash_compare, buddy_hash_cleanup);
	buddy_list_fn = buddy_list_file(priv->username);
    read_buddylist(priv, buddy_list_fn);
	free(buddy_list_fn);
	blist_init(priv);

	return (0);
}

void buddy_destroy(struct aim_priv *priv) {
	struct buddy_pref *pref = priv->buddy_pref;

	if (pref->group_list || pref->permit_list || pref->block_list) {
		char *buddy_list_fn = buddy_list_file(priv->username);

		buddy_save_list(priv, buddy_list_fn);
		free(buddy_list_fn);
	}

	hash_destroy(&pref->buddy_hash);

	dlist_destroy(pref->group_list, group_cleanup);
	dlist_destroy(pref->permit_list, generic_cleanup);
	dlist_destroy(pref->block_list, generic_cleanup);

	pref->group_list = NULL;
	pref->permit_list = NULL;
	pref->block_list = NULL;

	free(pref);
}

int read_buddylist(struct aim_priv *priv, const char *filename) {
	FILE *fp;
	char buf[1024];
	u_int32_t line = 0;
	struct group *cur_group = NULL;
	struct buddy_pref *pref = priv->buddy_pref;

	fp = fopen(filename, "r");
	if (fp == NULL) {
		if (errno != EEXIST)
			screen_status_msg("Can't open buddy list: %s: %s", filename, strerror(errno));
		return (-1);
	}

	pref->report_idle = opt_get_bool(OPT_REPORT_IDLE);
	pref->permdeny_mode = 1;

	while (fgets(buf, sizeof(buf), fp) != NULL) {
		char *p;
		char type;

		++line;
		p = strchr(buf, '\n');
		if (p == NULL) {
			screen_status_msg("Invalid buddy list data at line %u", line);
			fclose(fp);
			return (-1);
		}

		*p = '\0';
		p = buf;

		type = *p++;

		while (*p == ' ' || *p == '\t')
			p++;

		if (type == 'g') {
			struct group *group;

			group = xcalloc(1, sizeof(*group));
			group->name = xstrdup(p);
			group->refnum = pref->next_group_refnum++;
			pref->group_list = dlist_add_tail(pref->group_list, group);
			cur_group = group;
		} else if (type == 'b') {
			char *alias;
			struct buddy *buddy;

			if (cur_group == NULL) {
				screen_status_msg("Invalid buddy list data at line %u", line);
				fclose(fp);
			}
			alias = strchr(p, ':');
			if (alias != NULL)
				*alias++ = '\0';

			buddy = buddy_add(priv, p, cur_group);
			if (alias != NULL)
				buddy_alias(priv, buddy, alias);
		} else if (type == 'p') {
			pref->permit_list = dlist_add_tail(pref->permit_list, p);
		} else if (type == 'b') {
			pref->block_list = dlist_add_tail(pref->block_list, p);
		} else if (type == 'm') {
			u_int32_t permdeny_mode;
			char *end;

			permdeny_mode = strtoul(p, &end, 10);
			if (*end != '\0' || permdeny_mode > 5) {
				screen_status_msg(
					"%s: Invalid value for permit/deny mode at line %u", p, line);
				fclose(fp);
				return (-1);
			}
			pref->permdeny_mode = permdeny_mode;
		} else if (type == 'i') {
			u_int32_t report_idle_val;
			char *end;

			report_idle_val = strtoul(p, &end, 10);
			if (*end != '\0' || report_idle_val > 1) {
				screen_status_msg(
					"Invalid value for report idle mode at line %u", line);
				fclose(fp);
				return (-1);
			}
			pref->report_idle = report_idle_val;
		} else {
			screen_status_msg("Invalid buddy list data at line %u", line);
			fclose(fp);
			return (-1);
		}
	}
	
	fclose(fp);
	return (0);
}

int write_buddylist(struct aim_priv *priv, const char *filename) {
	FILE *fp;
	dlist_t *gcur;
	struct buddy_pref *pref = priv->buddy_pref;

	fp = fopen(filename, "w");
	if (fp == NULL) {
		screen_status_msg("Can't open buddy list file for writing: %s",
			strerror(errno));
		return (-1);
	}

	fprintf(fp, "m %u\n", pref->permdeny_mode);
	fprintf(fp, "i %u\n", pref->report_idle);

	for (gcur = pref->group_list ; gcur != NULL ; gcur = gcur->next) {
		struct group *gr = gcur->data;
		dlist_t *bcur;

		fprintf(fp, "g %s\n", gr->name);

		for (bcur = gr->member_list ; bcur != NULL ; bcur = bcur->next) {
			struct buddy *buddy = bcur->data;

			fprintf(fp, "b %s:%s\n", buddy->nname, buddy->name);
		}
	}

	for (gcur = pref->permit_list ; gcur != NULL ; gcur = gcur->next)
		fprintf(fp, "p %s\n", (char *) gcur->data);

	for (gcur = pref->block_list ; gcur != NULL ; gcur = gcur->next)
		fprintf(fp, "b %s\n", (char *) gcur->data);

	fclose(fp);
	return (0);
}

int buddy_save_list(struct aim_priv *priv, char *filename) {
	char *blist_fn_temp;
	size_t len; 

	len = strlen(filename) + sizeof("-TEMP");
	blist_fn_temp = xmalloc(len);
	snprintf(blist_fn_temp, len, "%s-TEMP", filename);

	if (write_buddylist(priv, blist_fn_temp) != 0) {
		free(blist_fn_temp);
		return (-1);
	}

	if (rename(blist_fn_temp, filename) != 0) {
		unlink(blist_fn_temp);
		free(blist_fn_temp);
		return (-1);
	}

	free(blist_fn_temp);
	return (0);
}

void buddy_update(	struct aim_priv *priv,
					struct buddy *buddy,
					aim_userinfo_t *userinfo)
{
	buddy->signon_time = userinfo->onlinesince;
	buddy->warn_level = (float) userinfo->warnlevel / 10;

	if (userinfo->present & AIM_USERINFO_PRESENT_FLAGS) {
		if (userinfo->flags & AIM_FLAG_UNCONFIRMED)
			buddy->type |= USER_TYPE_UNCONFIRMED;
		if (userinfo->flags & AIM_FLAG_FREE)
			buddy->type |= USER_TYPE_AIM;
		if (userinfo->flags & AIM_FLAG_AOL)
			buddy->type |= USER_TYPE_AOL;
		if (userinfo->flags & AIM_FLAG_WIRELESS)
			buddy->type |= USER_TYPE_WIRELESS;
		if (userinfo->flags & AIM_FLAG_ADMINISTRATOR)
			buddy->type |= USER_TYPE_ADMIN;
		if (userinfo->flags & AIM_FLAG_ACTIVEBUDDY)
			buddy->type |= USER_TYPE_ACTIVEBUDDY;
	}

	if (userinfo->present & AIM_USERINFO_PRESENT_SESSIONLEN)
		buddy->idle_time = userinfo->idletime;

	buddy->status = STATUS_ACTIVE;

	if (buddy->idle_time > 0)
		buddy->status = STATUS_IDLE;

	if (userinfo->flags & AIM_FLAG_AWAY)
		buddy->status = STATUS_AWAY;

	if (buddy->blist_line != NULL)
		blist_draw_buddy_label(priv->blist, buddy);
}

void buddy_online(	struct aim_priv *priv,
					struct buddy *buddy,
					aim_userinfo_t *userinfo)
{
	buddy->group->num_online++;
	buddy_update(priv, buddy, userinfo);
	blist_add(priv->blist, buddy);
}

void buddy_offline(struct aim_priv *priv, struct buddy *buddy) {
	buddy->status = STATUS_OFFLINE;
	buddy->signon_time = 0;
	buddy->idle_time = 0;
	buddy->warn_level = 0;
	buddy->group->num_online--;
	blist_del(priv->blist, buddy);
}

int buddy_add_block(struct aim_priv *priv, char *screen_name) {
	dlist_t *node;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	normalize(screen_name, nname, sizeof(nname));

	node = dlist_find(pref->block_list, nname, buddy_rights_compare);
	if (node != NULL)
		return (-1);

	pref->block_list = dlist_add_tail(pref->block_list, xstrdup(nname));
	aim_add_block(priv, nname);
	return (0);
}

int buddy_remove_block(struct aim_priv *priv, char *screen_name) {
	dlist_t *node;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	normalize(screen_name, nname, sizeof(nname));

	node = dlist_find(pref->block_list, nname, buddy_rights_compare);
	if (node == NULL)
		return (-1);

	aim_remove_block(priv, nname);
	free(node->data);
	pref->block_list = dlist_remove(pref->block_list, node);
	return (0);
}

void buddy_print_list(struct aim_priv *priv) {
	struct buddy_pref *pref = priv->buddy_pref;
	dlist_t *gcur = pref->group_list;

	while (gcur != NULL) {
		struct group *gr = gcur->data;
		dlist_t *bcur = gr->member_list;

		screen_status_msg("+ %s (%u/%u)",
			gr->name, gr->num_online, gr->num_members);

		while (bcur != NULL) {
			struct buddy *buddy = bcur->data;

			if (buddy->status != STATUS_OFFLINE) {
				screen_status_msg("  - %s (online) [i: %u, w: %u]",
					buddy->name, buddy->idle_time, buddy->warn_level);
			} else {
				screen_status_msg("  - %s (offline)", buddy->name);
			}

			bcur = bcur->next;
		}

		gcur = gcur->next;
	}
}

void buddy_print_block(struct aim_priv *priv) {
	struct buddy_pref *pref = priv->buddy_pref;
	dlist_t *cur = pref->block_list;

	if (cur == NULL) {
		screen_status_msg("The blocked users list is empty");
		return;
	}

	screen_status_msg("Blocked users list:");

	while (cur != NULL) {
		screen_status_msg(" - %s", (char *) cur->data);
		cur = cur->next;
	}
}

int buddy_add_permit(struct aim_priv *priv, char *screen_name) {
	dlist_t *node;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	normalize(screen_name, nname, sizeof(nname));

	node = dlist_find(pref->permit_list, nname, buddy_rights_compare);
	if (node != NULL)
		return (-1);

	aim_add_permit(priv, nname);
	pref->permit_list = dlist_add_tail(pref->permit_list, xstrdup(nname));
	return (0);
}

int buddy_remove_permit(struct aim_priv *priv, char *screen_name) {
	dlist_t *node;
	char nname[NUSER_LEN];
	struct buddy_pref *pref = priv->buddy_pref;

	normalize(screen_name, nname, sizeof(nname));

	node = dlist_find(pref->permit_list, nname, buddy_rights_compare);
	if (node == NULL)
		return (-1);

	aim_remove_permit(priv, nname);
	free(node->data);
	pref->permit_list = dlist_remove(pref->permit_list, node);
	return (0);
}

void buddy_print_permit(struct aim_priv *priv) {
	struct buddy_pref *pref = priv->buddy_pref;
	dlist_t *cur = pref->permit_list;

	if (cur == NULL) {
		screen_status_msg("Permitted users list is empty");
		return;
	}

	screen_status_msg("Permitted users list:");

	while (cur != NULL) {
		screen_status_msg(" - %s", (char *) cur->data);
		cur = cur->next;
	}
}

void buddy_update_idle(struct aim_priv *priv) {
	struct buddy_pref *pref = priv->buddy_pref;
	dlist_t *gcur = pref->group_list;

	while (gcur != NULL) {
		struct group *gr = gcur->data;
		dlist_t *bcur = gr->member_list;

		while (bcur != NULL) {
			struct buddy *buddy = bcur->data;

			if (buddy->idle_time > 0) {
				buddy->idle_time++;
				if (buddy->blist_line != NULL)
					blist_draw_buddy_label(priv->blist, buddy);
			}

			bcur = bcur->next;
		}

		gcur = gcur->next;
	}
}
