/*
** pork_misc.c - miscellaneous functions.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>

#include <pork_util.h>
#include <pork_list.h>
#include <pork_input.h>
#include <pork_misc.h>

/*
** Initialize the colors that we're going to be using.
*/

void init_colors(void) {
	start_color();

	init_pair(0, COLOR_BLACK, COLOR_BLACK);
	init_pair(1, COLOR_RED,	COLOR_BLACK);
	init_pair(2, COLOR_GREEN, COLOR_BLACK);
	init_pair(3, COLOR_YELLOW, COLOR_BLACK);
	init_pair(4, COLOR_BLUE, COLOR_BLACK);
	init_pair(5, COLOR_MAGENTA, COLOR_BLACK);
	init_pair(6, COLOR_CYAN, COLOR_BLACK);
	init_pair(7, COLOR_WHITE, COLOR_BLACK);

	init_pair(8, COLOR_GREEN, COLOR_GREEN);
	init_pair(9, COLOR_RED, COLOR_RED);
	init_pair(10, COLOR_YELLOW, COLOR_YELLOW);
	init_pair(11, COLOR_YELLOW, COLOR_BLUE);
}

/*
** Sets default options for windows.
*/ 

void set_default_win_opts(WINDOW *win) {
	leaveok(win, FALSE);
	scrollok(win, FALSE);
	meta(win, TRUE);
	keypad(win, TRUE);
	idlok(win, TRUE);
}

/*
** Setup the ncurses environment.
*/

int initialize_environment(void) {
	WINDOW *win;

	win = initscr();
	if (win == NULL)
		return (-1);

	halfdelay(10);
	noecho();
	raw();
	set_default_win_opts(win);

	return (0);
}

int wgetinput(WINDOW *win) {
	int c = wgetch(win);

	if (c == 0x1b) {
		c = wgetch(win);
		if (c == -1)
			return (-1);
		c |= ALT_MASK;
	}

	return (c);
}

/*
** I define what I call "cstrings" for
** use in this program.
**
** ncurses uses a data type called "chtype" to hold
** a character with attribute set (f.e. having color, being bold,
** underlined, etc).
**
** I define an array of these characters that's terminated with a chtype
** character having the value of zero as a "cstr"
*/

/*
** Return the length of the cstring pointed
** by "ch"
*/

inline size_t cstrlen(chtype *ch) {
	size_t i = 0;

	while (*ch++ != 0)
		++i;

	return (i);
}

/*
** Convert the cstring specified by "cstr"
** to its ASCII representation.  This will result
** in the loss of all attributes of the characters
** that comprise it.
*/

char *cstr_to_plaintext(chtype *cstr, size_t n) {
	char *str = xmalloc(n + 1);
	size_t i;

	for (i = 0 ; i < n ; i++) {
		str[i] = chtype_get(cstr[i]);
		if (cstr[i] == 0)
			break;
	}

	return (str);
}

/*
** The opposite as above.  The characters that comprise
** the cstring that will be constructed will have no
** special attributes set.
*/

int plaintext_to_cstr(const char *str, chtype *ch, size_t len) {
	u_int32_t i;

	len--;
	for (i = 0 ; i < len && str[i] != '\0' ; i++)
		ch[i] = str[i];

	ch[i] = 0;
	return (i);
}

/*
** Write the cstring pointed to by "ch"
** to the screen at the current cursor position.
*/

inline void wputstr(WINDOW *win, chtype *ch) {
	chtype *p = ch;

	while (*p != 0)
		waddch(win, *p++);
}

/*
** Write the cstring pointed to by "ch"
** to the screen at the position (x, y).
*/

inline void mvwputstr(WINDOW *win, int y, int x, chtype *ch) {
	wmove(win, y, x);
	wputstr(win, ch);
}

/*
** Write the first n chtype chars of the cstring pointed to by "ch"
** to the screen at the current cursor position.
*/

inline void wputnstr(WINDOW *win, chtype *ch, size_t n) {
	size_t i;

	for (i = 0 ; i < n && ch[i] != 0 ; i++)
		waddch(win, ch[i]);
}

/*
** Write the first n chtype chars of the cstring pointed to by "ch"
** to the screen at position (x, y).
*/

inline void mvwputnstr(WINDOW *win, int y, int x, chtype *ch, size_t n) {
	wmove(win, y, x);
	wputnstr(win, ch, n);
}

/*
** "Normalize" the string "str"
**
** AIM screen names need to be normalized
** before they're sent to the server.
**
** What this means is converting alphabetic characters
** to lowercase and removing any space characters.
*/

void normalize(const char *str, char *buf, size_t len) {
	size_t i;

	len--;
	for (i = 0 ; *str != '\0' && i < len ; str++) {
		if (*str == ' ')
			continue;

		if (isalpha(*str))
			buf[i++] = tolower(*str);
		else
			buf[i++] = *str;
	}

	buf[i] = '\0';
}

/*
** Strip any html tags from "str"
** Replace common "&*;" forms with their ascii representation.
*/

char *strip_html(char *str) {
	char *p;
	char *buf = xmalloc(strlen(str) + 1);
	char *ret = buf;
	int ignore = 0;
	char *last_tag = NULL;

	for (p = str ; *p != '\0' ; p++) {
		if (ignore == 1) {
			if (*p == '>') {
				ignore = 0;

				if (last_tag != NULL) {
					if (!strncasecmp(last_tag, "<br>", 4) ||
						!strncasecmp(last_tag, "<p>", 3))
					{
						*buf++ = '\n';
					}

					last_tag = NULL;
				}
			} else if (*p == '<')
				last_tag = p;

			continue;
		}

		if (*p == '<') {
			ignore = 1;
			last_tag = p;
			continue;
		} else if (*p == '&') {
			char *q = strchr(p, ';');

			p++;
			if (q != NULL) {
				*q = '\0';

				if (!strcasecmp(p, "nbsp"))
					*buf++ = ' ';
				else if (!strcasecmp(p, "lt"))
					*buf++ = '<';
				else if (!strcasecmp(p, "gt"))
					*buf++ = '>';
				else if (!strcasecmp(p, "amp"))
					*buf++ = '&';
				else if (!strcasecmp(p, "quot"))
					*buf++ = '"';
				else if (!strcasecmp(p, "apos"))
					*buf++ = '\'';
				else if (!strcasecmp(p, "hyphen"))
					*buf++ = '-';
				else if (!strcasecmp(p, "comma"))
					*buf++ = ',';
				else if (!strcasecmp(p, "excl"))
					*buf++ = '!';
				else if (!strcasecmp(p, "iexcl"))
					*buf++ = 'i';
				else if (!strcasecmp(p, "percnt"))
					*buf++ = '%';
				else if (!strcasecmp(p, "num"))
					*buf++ = '#';
				else if (!strcasecmp(p, "ast") || !strcasecmp(p, "mul"))
					*buf++ = '*';
				else if (!strcasecmp(p, "commat"))
					*buf++ = '@';
				else if (!strcasecmp(p, "lsqb"))
					*buf++ = '[';
				else if (!strcasecmp(p, "rsqb"))
					*buf++ = ']';
				else if (!strcasecmp(p, "bsol"))
					*buf++ = '\\';
				else if (!strcasecmp(p, "lcub"))
					*buf++ = '{';
				else if (!strcasecmp(p, "plus"))
					*buf++ = '+';
				else if (!strcasecmp(p, "divide"))
					*buf++ = '/';
				else if (!strcasecmp(p, "rcub"))
					*buf++ = '}';
				else if (!strcasecmp(p, "colon"))
					*buf++ = ':';
				else if (!strcasecmp(p, "period"))
					*buf++ = '.';
				else if (!strcasecmp(p, "semi"))
					*buf++ = ';';
				else if (!strcasecmp(p, "copy")) {
					strcpy(buf, "(C)");
					buf += 3;
				} else if (!strcasecmp(p, "trade")) {
					strcpy(buf, "(TM)");
					buf += 3;
				} else if (!strcasecmp(p, "reg")) {
					strcpy(buf, "(R)");
					buf += 3;
				} else {
					*buf++ = *p;
					continue;
				}

				p = q;
				continue;
			}
		} else if (*p == '\r')
			*buf++ = '\n';
		else
			*buf++ = *p;
	}

	*buf++ = '\0';

	return (ret);
}

/*
** Convert "timespec" (in UNIX time) to a string of the
** form "DAYSd HOURSh MINm"
*/ 

void time_to_str(u_int32_t timespec, char *buf, size_t len) {
	u_int32_t days;
	u_int32_t hours;
	u_int32_t min;
	u_int32_t i = 0;

	if (timespec == 0) {
		snprintf(buf, len, "0");
		return;
	}

	days = timespec / (60 * 24);
	hours = (timespec % (60 * 24)) / 60;
	min = timespec - (days * (60 * 24)) - (hours * 60);

	if (days > 0) {
		int ret = snprintf(buf, len, "%ud ", days);
		len -= ret;
		i += ret;
	}

	if (hours > 0) {
		int ret = snprintf(&buf[i], len, "%uh ", hours);
		len -= ret;
		i += ret;
	}

	if (min > 0) {
		int ret = snprintf(&buf[i], len, "%um", min);
		len -= ret;
		i += ret;
	}
}

/*
** This is a utility function for constructing output lines
** that are used to present users' profiles.
**
** This is slightly out of place here.  But, hey, why not.
*/

void chstr_color_info(chtype *ch) {
	size_t i = 1;

	if (ch == NULL || ch[0] == '\0')
		return;

	ch[0] = chtype_get(ch[0]) | ATTR_INFO_FIRST;

	while (1) {
		char c = chtype_get(ch[i]);

		if (c == '\0' || c == ':')
			break;

		ch[i++] = c | ATTR_INFO;
	}
}
