/*
** pork_screen_cmd.c - screen management.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** The functions in this file are mostly wrappers around
** functions that do the actual work.  The difference is
** these functions do any necessary updating of the contents
** of the screen (i.e. refreshing, redrawing, etc).
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>

#include <pork.h>

inline void screen_draw_cursor(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;

	wmove(scr->status_bar, STATUS_ROWS - 1,
		input_get_cursor_pos(imwindow->input));
}

void screen_redraw_input(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;
	char *input_line;

	wmove(scr->status_bar, STATUS_ROWS - 1, 0);
	wclrtoeol(scr->status_bar);

	input_line = input_partial(imwindow->input);
	if (input_line == imwindow->input->input_buf &&
		imwindow->input->prompt != NULL)
	{
		waddstr(scr->status_bar, imwindow->input->prompt);
	}

	waddnstr(scr->status_bar,
		input_partial(imwindow->input), imwindow->input->width);

	screen_draw_cursor(scr);
	wnoutrefresh(scr->status_bar);
}

void screen_window_swap(struct screen *scr, dlist_t *new_cur) {
	struct imwindow *imwindow = NULL;
	u_int32_t last_own_input = 0;
	u_int32_t cur_own_input;
	struct aim_priv *priv;

	if (scr->cur_window != NULL) {
		imwindow = scr->cur_window->data;

		last_own_input = wopt_get_bool(imwindow->opts, WOPT_PRIVATE_INPUT);
		imwindow->swindow.visible = FALSE;
	}

	scr->cur_window = new_cur;

	imwindow = scr->cur_window->data;
	imwindow->swindow.visible = TRUE;
	imwindow->swindow.dirty = FALSE;
	cur_own_input = wopt_get_bool(imwindow->opts, WOPT_PRIVATE_INPUT);

	/*
	** If the current window and the one that we're going to switch
	** to don't share an input buffer, history, etc, the input
	** line must be refreshed.
	*/

	if (cur_own_input || last_own_input)
		screen_redraw_input(scr);

	status_draw(scr, imwindow->priv);

	priv = imwindow->priv;

	if (imwindow->blist_visible && priv->blist->win != NULL) {
		redrawwin(priv->blist->win);
		wnoutrefresh(priv->blist->win);
	}

	redrawwin(imwindow->swindow.win);
	imwindow_refresh(imwindow);
}

inline int screen_goto_window(struct screen *scr, u_int32_t refnum) {
	dlist_t *cur = screen_find_refnum(scr, refnum);

	if (cur == NULL)
		return (-1);

	screen_window_swap(scr, cur);
	return (0);
}

inline void screen_refresh(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;
	struct aim_priv *priv = imwindow->priv;
	int cur_old = curs_set(0);

	wclear(curscr);
	status_draw(scr, imwindow->priv);
	redrawwin(imwindow->swindow.win);

	if (priv->blist != NULL) {
		if (imwindow->blist_visible && priv->blist->win != NULL)
			redrawwin(priv->blist->win);
		blist_refresh(priv->blist, imwindow);
	}

	screen_redraw_input(scr);
	imwindow_refresh(imwindow);
	curs_set(cur_old);
}

inline void screen_clear(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;

	swindow_clear(&imwindow->swindow);
	imwindow_refresh(imwindow);
}

inline void screen_erase(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;

	swindow_erase(&imwindow->swindow);
	imwindow_refresh(imwindow);
}

inline void screen_scroll_pgup(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;

	swindow_scroll_by(&imwindow->swindow, -imwindow->swindow.rows);
	imwindow_refresh(imwindow);
}

void screen_scroll_by(struct screen *scr, int lines) {
	struct imwindow *imwindow = scr->cur_window->data;

	swindow_scroll_by(&imwindow->swindow, lines);
	imwindow_refresh(imwindow);
}

inline void screen_scroll_up(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;

	swindow_scroll_up(&imwindow->swindow);
	imwindow_refresh(imwindow);
}

inline void screen_scroll_down(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;

	swindow_scroll_down(&imwindow->swindow);
	imwindow_refresh(imwindow);
}

inline void screen_scroll_to_end(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;

	swindow_scroll_to_end(&imwindow->swindow);
	imwindow_refresh(imwindow);
}

inline void screen_scroll_to_start(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;

	swindow_scroll_to_start(&imwindow->swindow);
	imwindow_refresh(imwindow);
}

inline void screen_scroll_pgdown(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;

	swindow_scroll_by(&imwindow->swindow, imwindow->swindow.rows);
	imwindow_refresh(imwindow);
}

struct imwindow *screen_new_window(	struct screen *scr,
									struct aim_priv *priv,
									char *target)
{
	u_int32_t refnum = screen_get_new_refnum(scr);
	struct imwindow *imwindow;

	imwindow = imwindow_new(scr->rows, scr->cols,
		refnum, TYPE_IM, priv, target);
	if (imwindow == NULL)
		return (NULL);

	screen_add_window(scr, imwindow);
	status_draw(scr, imwindow->priv);

	imwindow_refresh(imwindow);
	return (imwindow);
}

inline void screen_cycle_fwd(struct screen *scr) {
	screen_window_swap(scr, scr->cur_window->next);
}

inline void screen_cycle_bak(struct screen *scr) {
	screen_window_swap(scr, scr->cur_window->prev);
}

int screen_close_window(struct screen *scr) {
	struct imwindow *imwindow = scr->cur_window->data;
	dlist_t *cur_window = scr->cur_window;

	/*
	** The status window can't be closed.
	*/

	if (imwindow->type == TYPE_STATUS)
		return (-1);

	screen_cycle_fwd(scr);
	screen_window_list_remove(scr, cur_window);
	imwindow_destroy(imwindow);
	free(cur_window);

	return (0);
}

int screen_set_quiet(struct screen *scr, int status) {
	int ret = scr->quiet;

	scr->quiet = status;
	return (ret);
}

/*
** Print a status message to the current window.
*/

void screen_status_msg(char *fmt, ...) {
	static char buf[4096];
	static chtype ch[4096];
	size_t buflen = sizeof(buf);
	size_t banner_len;
	va_list ap;
	char *banner;
	struct imwindow *imwindow = screen.cur_window->data;

	if (screen.quiet == 1)
		return;

	banner = opt_get_str(OPT_BANNER);
	banner_len = strlen(banner);

	xstrncpy(buf, banner, sizeof(buf));
	buflen -= banner_len;

	va_start(ap, fmt);
	vsnprintf(&buf[banner_len], buflen, fmt, ap);
	va_end(ap);

	plaintext_to_cstr(buf, ch, sizeof(ch));
	swindow_add(&imwindow->swindow, ch, 0);
	imwindow_refresh(imwindow);
}

void screen_buddy_status_msg(char *dest, int timestamp, char *fmt, ...) {
	char buf[4096];
	chtype ch[4096];
	size_t buflen = sizeof(buf);
	size_t banner_len;
	va_list ap;
	char *banner;
	struct imwindow *imwindow;

	imwindow = imwindow_find(dest);
	if (imwindow == NULL)
		imwindow = screen.cur_window->data;

	banner = opt_get_str(OPT_BANNER);
	banner_len = strlen(banner);

	xstrncpy(buf, banner, sizeof(buf));
	buflen -= banner_len;

	va_start(ap, fmt);
	vsnprintf(&buf[banner_len], buflen, fmt, ap);
	va_end(ap);

	plaintext_to_cstr(buf, ch, sizeof(ch));
	swindow_add(&imwindow->swindow, ch, timestamp);
	imwindow_refresh(imwindow);
}

/*
** Finish this.
*/

void screen_print_info(	char *user,
						char *profile,
						char *away_msg,
						u_int32_t idle_time,
						u_int32_t online_since,
						u_int32_t member_since,
						int print_all)
{
	struct imwindow *imwindow = screen.cur_window->data;
	char buf[2048];
	chtype ch[2048];

	if (print_all) {
		snprintf(buf, sizeof(buf), "user: %s", user);
		plaintext_to_cstr(buf, ch, sizeof(ch));
		chstr_color_info(ch);
		swindow_add(&imwindow->swindow, ch, 0);

		if (idle_time > 0) {
			char time_buf[64];

			time_to_str(idle_time, time_buf, sizeof(time_buf));

			snprintf(buf, sizeof(buf), "idle: %s", time_buf);
			plaintext_to_cstr(buf, ch, sizeof(ch));
			chstr_color_info(ch);
			swindow_add(&imwindow->swindow, ch, 0);
		}

		if (online_since) {
			snprintf(buf, sizeof(buf), "signon: %s",
				asctime(localtime((time_t *) &online_since)));
			plaintext_to_cstr(buf, ch, sizeof(ch));
			chstr_color_info(ch);
			swindow_add(&imwindow->swindow, ch, 0);
		}

		if (member_since) {
			snprintf(buf, sizeof(buf), "member since: %s",
				asctime(localtime((time_t *) &online_since)));
			plaintext_to_cstr(buf, ch, sizeof(ch));
			chstr_color_info(ch);
			swindow_add(&imwindow->swindow, ch, 0);
		}

		if (profile) {
			snprintf(buf, sizeof(buf), "profile:");
			plaintext_to_cstr(buf, ch, sizeof(ch));
			chstr_color_info(ch);
			swindow_add(&imwindow->swindow, ch, 0);
			screen_print_multiline(imwindow, profile, 0);
		}
	}

	if (away_msg) {
		snprintf(buf, sizeof(buf), "away message:");
		plaintext_to_cstr(buf, ch, sizeof(ch));
		chstr_color_info(ch);
		swindow_add(&imwindow->swindow, ch, 0);
		screen_print_multiline(imwindow, away_msg, 0);
	}

	imwindow_refresh(imwindow);
}

/*
** Print out a string that may have newlines
** embedded in it to the specified window.
*/

void screen_print_multiline(struct imwindow *win,
							char *msg,
							int timestamp)
{
	size_t len = strlen(msg) + 8;
	chtype *buf = xmalloc(sizeof(chtype) * len);
	char *p;

	p = msg;
	while (p != NULL) {
		char *next = strchr(p, '\n');
		if (next != NULL)
			*next++ = '\0';

		buf[0] = buf[1] = ' ';
		plaintext_to_cstr(p, &buf[2], len - 2);
		swindow_add(&win->swindow, buf, timestamp);
		p = next;
	}

	free(buf);
}

int screen_get_passwd(char *buf, size_t len) {
	int ret;

	buf[0] = '\0';

	wmove(screen.status_bar, 1, 0);
	wclrtoeol(screen.status_bar);

	waddstr(screen.status_bar, "Password: ");
	wrefresh(screen.status_bar);

	ret = wgetnstr(screen.status_bar, buf, len);
	return (ret);
}

/*
** Just in case anything needs to be done here in the future..
*/

inline void screen_doupdate(void) {
	int cur_old = curs_set(0);
	doupdate();
	curs_set(cur_old);
}
